@@ -1,6 +1,6 @@
 /*****************************************************************************
  *  Written by Chris Dunlap <cdunlap@llnl.gov>.
- *  Copyright (C) 2007-2017 Lawrence Livermore National Security, LLC.
+ *  Copyright (C) 2007-2020 Lawrence Livermore National Security, LLC.
  *  Copyright (C) 2002-2007 The Regents of the University of California.
  *  UCRL-CODE-155910.
  *
@@ -35,6 +35,7 @@
 #include <ctype.h>
 #include <errno.h>
 #include <limits.h>
+#include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -46,6 +47,7 @@
 #include "query.h"
 #include "read.h"
 #include "version.h"
+#include "xsignal.h"
 
 
 /*****************************************************************************
@@ -126,9 +128,8 @@ main (int argc, char *argv[])
     conf_t      conf;
     const char *p;
 
-    if (posignal (SIGPIPE, SIG_IGN) == SIG_ERR) {
-        log_err (EMUNGE_SNAFU, LOG_ERR, "Failed to ignore signal=%d", SIGPIPE);
-    }
+    xsignal_ignore (SIGHUP);
+    xsignal_ignore (SIGPIPE);
     log_open_file (stderr, argv[0], LOG_INFO, LOG_OPT_PRIORITY);
     conf = create_conf ();
     parse_cmdline (conf, argc, argv);
@@ -151,6 +152,7 @@ main (int argc, char *argv[])
     display_cred (conf);
 
     destroy_conf (conf);
+    log_close_file ();
     exit (EMUNGE_SUCCESS);
 }
 
@@ -357,16 +359,21 @@ parse_cmdline (conf_t conf, int argc, char **argv)
             case 't':
                 errno = 0;
                 l = strtol (optarg, &p, 10);
-                if ((optarg == p) || (*p != '\0')) {
+                if ((optarg == p) || (*p != '\0') || (l < -1)) {
                     log_err (EMUNGE_SNAFU, LOG_ERR,
                         "Invalid time-to-live '%s'", optarg);
                 }
-                if (((errno == ERANGE) && (l == LONG_MAX)) || (l > INT_MAX)) {
+                if ((errno == ERANGE) && (l == LONG_MAX)) {
                     log_err (EMUNGE_SNAFU, LOG_ERR,
-                        "Exceeded maximum time-to-live of %d seconds",
-                        INT_MAX);
+                        "Overflowed maximum time-to-live of %ld seconds",
+                        LONG_MAX);
                 }
-                if (l < 0) {
+                if (l > UINT_MAX) {
+                    log_err (EMUNGE_SNAFU, LOG_ERR,
+                        "Exceeded maximum time-to-live of %u seconds",
+                        UINT_MAX);
+                }
+                if (l == -1) {
                     l = MUNGE_TTL_MAXIMUM;
                 }
                 e = munge_ctx_set (conf->ctx, MUNGE_OPT_TTL, (int) l);
