@@ -1,6 +1,6 @@
 /*****************************************************************************
  *  Written by Chris Dunlap <cdunlap@llnl.gov>.
- *  Copyright (C) 2007-2017 Lawrence Livermore National Security, LLC.
+ *  Copyright (C) 2007-2020 Lawrence Livermore National Security, LLC.
  *  Copyright (C) 2002-2007 The Regents of the University of California.
  *  UCRL-CODE-155910.
  *
@@ -804,15 +804,21 @@ dec_unpack_inner (munge_cred_t c)
     /*
      *  Unpack the origin IP address.
      */
-    if (m->addr_len != sizeof (m->addr)) {
-        return (m_msg_set_err (m, EMUNGE_BAD_CRED,
-            strdup ("Invalid origin IP addr length")));
-    }
     if (m->addr_len > len) {
         return (m_msg_set_err (m, EMUNGE_BAD_CRED,
             strdup ("Truncated origin IP addr")));
     }
-    memcpy (&m->addr, p, m->addr_len);
+    else if (m->addr_len == 4) {
+        assert (sizeof (m->addr) == 4);
+        memcpy (&m->addr, p, m->addr_len);
+    }
+    else if (m->addr_len == 0) {
+        memset (&m->addr, 0, sizeof (m->addr));
+    }
+    else {
+        return (m_msg_set_err (m, EMUNGE_BAD_CRED,
+            strdup ("Invalid origin IP addr length")));
+    }
     p += m->addr_len;
     len -= m->addr_len;
     /*
@@ -1012,7 +1018,7 @@ dec_validate_replay (munge_cred_t c)
         if ((conf->got_socket_retry)
                 && (m->retry > 0)
                 && (m->retry <= MUNGE_SOCKET_RETRY_ATTEMPTS)) {
-            log_msg (LOG_NOTICE,
+            log_msg (LOG_INFO,
                 "Allowed credential replay for client UID=%u GID=%u",
                 (unsigned int) m->client_uid, (unsigned int) m->client_gid);
             return (0);
