@@ -1,6 +1,6 @@
 /*****************************************************************************
  *  Written by Chris Dunlap <cdunlap@llnl.gov>.
- *  Copyright (C) 2007-2017 Lawrence Livermore National Security, LLC.
+ *  Copyright (C) 2007-2020 Lawrence Livermore National Security, LLC.
  *  Copyright (C) 2002-2007 The Regents of the University of California.
  *  UCRL-CODE-155910.
  *
@@ -36,16 +36,12 @@
 #include <munge.h>
 #include <stdint.h>
 #include <string.h>
-#if HAVE_SYS_RANDOM_H
-#include <sys/random.h>
-#endif /* HAVE_SYS_RANDOM_H */
 #include <sys/stat.h>
-#include <sys/time.h>
-#include <time.h>
 #include <unistd.h>
 #include "common.h"
 #include "conf.h"
 #include "crypto.h"
+#include "entropy.h"
 #include "log.h"
 #include "munge_defs.h"
 #include "path.h"
@@ -57,10 +53,6 @@
  *  Constants
  *****************************************************************************/
 
-/*  String specifying the pathname of the random number source device.
- */
-#define RANDOM_SOURCE_PATH              "/dev/urandom"
-
 /*  Integer for the number of bytes to read from the random number source
  *    device when seeding the PRNG entropy pool.
  *  Note that an upper limit of 256 bytes is imposed when using either
@@ -148,14 +140,8 @@ random_init (const char *seed_path)
         num_bytes_entropy += n;
     }
     if (num_bytes_entropy < RANDOM_BYTES_MIN) {
-        if (!conf->got_force) {
-            log_err (EMUNGE_SNAFU, LOG_ERR,
-                    "Failed to seed PRNG with sufficient entropy");
-        }
-        else {
-            log_msg (LOG_WARNING,
-                    "Failed to seed PRNG with sufficient entropy");
-        }
+        log_err_or_warn (conf->got_force,
+                "Failed to seed PRNG with sufficient entropy");
     }
     /*  Compute the initial time interval for stirring the entropy pool.
      *  If the desired amount of entropy is not available, increase the
@@ -247,98 +233,21 @@ _random_read_entropy_from_kernel (void)
 /*  Reads entropy from the kernel's CSPRNG.
  *  Returns the number of bytes of entropy added, or -1 on error.
  */
-    int            n = -1;
-    int            fd;
-    size_t         len;
+    int            n;
     const char    *src;
-    struct stat    st;
     unsigned char  buf [RANDOM_SOURCE_BYTES];
 
-#if HAVE_GETRANDOM
-    /*
-     * If the urandom source has been initialized, reads of up to 256 bytes
-     *   will always return as many bytes as requested and not be interrupted
-     *   by signals.  No such guarantees apply for larger buffer sizes.
-     */
-    src = "getrandom()";
-    len = MIN(256, sizeof (buf));
-retry_getrandom:
-    n = getrandom (buf, len, 0);
-    if (n < 0) {
-        if (errno == EINTR) {
-            goto retry_getrandom;
-        }
-        log_msg (LOG_WARNING, "Failed to fill buffer via getrandom(): %s",
-                strerror (errno));
-    }
-#elif HAVE_GETENTROPY
-    /*
-     *  The maximum buffer size permitted is 256 bytes.
-     */
-    src = "getentropy()";
-    len = MIN(256, sizeof (buf));
-    n = getentropy (buf, len);
-    if (n < 0) {
-        log_msg (LOG_WARNING, "Failed to fill buffer via getentropy(): %s",
-                strerror (errno));
-    }
-    else if (n == 0) {
-        n = len;
-    }
-#endif /* HAVE_GETENTROPY */
-
-    if (n < 0) {
-retry_open:
-        fd = open (RANDOM_SOURCE_PATH, O_RDONLY | O_NONBLOCK);
-        if (fd < 0) {
-            if (errno == EINTR) {
-                goto retry_open;
-            }
-            log_msg (LOG_WARNING, "Failed to open \"%s\": %s",
-                    RANDOM_SOURCE_PATH, strerror (errno));
-        }
-        else {
-            if (fstat (fd, &st) < 0) {
-                log_msg (LOG_WARNING, "Failed to stat \"%s\": %s",
-                        RANDOM_SOURCE_PATH, strerror (errno));
-            }
-            else if (!S_ISCHR (st.st_mode)) {
-                log_msg (LOG_WARNING, "Failed to validate \"%s\": "
-                        "not a character device (mode=0x%x)",
-                        RANDOM_SOURCE_PATH, (st.st_mode & S_IFMT));
-            }
-            else {
-                src = "\"" RANDOM_SOURCE_PATH "\"";
-                len = sizeof (buf);
-                n = fd_read_n (fd, buf, len);
-                if (n < 0) {
-                    log_msg (LOG_WARNING, "Failed to read from \"%s\": %s",
-                            RANDOM_SOURCE_PATH, strerror (errno));
-                }
-            }
-            if (close (fd) < 0) {
-                log_msg (LOG_WARNING, "Failed to close \"%s\": %s",
-                        RANDOM_SOURCE_PATH, strerror (errno));
-            }
-        }
-    }
-
+    n = entropy_read (buf, sizeof (buf), &src);
     if (n > 0) {
-        if (n > sizeof (buf)) {
-            log_msg (LOG_WARNING, "Ignoring entropy from \"%s\":"
-                    " read %d byte%s into a %d-byte buffer",
-                    RANDOM_SOURCE_PATH, n, (n == 1 ? "" : "s"), sizeof (buf));
-            n = 0;
-        }
-        else if (_random_check_entropy (buf, n) < 0) {
-            log_msg (LOG_WARNING, "Ignoring entropy from \"%s\":"
-                    " does not appear random", RANDOM_SOURCE_PATH);
+        if (_random_check_entropy (buf, n) < 0) {
             n = 0;
+            log_msg (LOG_WARNING,
+                    "Ignoring entropy from kernel: does not appear random");
         }
         else {
             _random_add (buf, n);
-            log_msg (LOG_INFO, "PRNG seeded with %d byte%s from %s",
-                    n, (n == 1 ? "" : "s"), src);
+            log_msg (LOG_INFO, "Seeded PRNG with %d byte%s from %s",
+                    n, (n == 1 ? "" : "s"), (src != NULL) ? src : "???");
         }
     }
     return (n);
@@ -371,11 +280,9 @@ _random_read_entropy_from_file (const char *path)
         log_err (EMUNGE_SNAFU, LOG_ERR,
                 "Failed to check PRNG seed dir \"%s\": %s", dir, ebuf);
     }
-    else if ((n == 0) && (!conf->got_force)) {
-        log_err (EMUNGE_SNAFU, LOG_ERR, "PRNG seed dir is insecure: %s", ebuf);
-    }
     else if (n == 0) {
-        log_msg (LOG_WARNING, "PRNG seed dir is insecure: %s", ebuf);
+        log_err_or_warn (conf->got_force,
+                "PRNG seed dir is insecure: %s", ebuf);
     }
     else {
         is_path_secure = 1;
@@ -404,18 +311,14 @@ _random_read_entropy_from_process (void)
 /*  Reads entropy from sources related to the process.
  *  Returns the number of bytes of entropy added, or -1 on error.
  */
-    pid_t  pid;
-    time_t now;
+    unsigned buf;
+    int      n = 0;
 
-    pid = getpid ();
-    _random_add (&pid, sizeof (pid));
-
-    if (time (&now) != (time_t) -1) {
-        _random_add (&now, sizeof (now));
+    if (entropy_read_uint (&buf) != -1) {
+        _random_add (&buf, sizeof (buf));
+        n += sizeof (buf);
     }
-    /*  Since these sources do not provide much entropy, return 0.
-     */
-    return (0);
+    return (n);
 }
 
 
@@ -445,7 +348,8 @@ _random_read_seed (const char *path, int num_bytes)
     is_symlink = (lstat (path, &st) == 0) ? S_ISLNK (st.st_mode) : 0;
     if (is_symlink) {
         log_msg (LOG_WARNING,
-                "Ignoring PRNG seed \"%s\": symbolic link not allowed", path);
+                "Ignoring PRNG seed \"%s\": must not be a symbolic link",
+                path);
         return (-1);
     }
 
@@ -470,20 +374,23 @@ retry_open:
     }
     else if (!S_ISREG (st.st_mode)) {
         log_msg (LOG_WARNING,
-                "Ignoring PRNG seed \"%s\": not a regular file (mode=0x%x)",
-                path, (st.st_mode & S_IFMT));
+                "Ignoring PRNG seed \"%s\": must be a regular file "
+                "(type=%07o)", path, (st.st_mode & S_IFMT));
     }
     else if (st.st_uid != geteuid ()) {
-        log_msg (LOG_WARNING, "Ignoring PRNG seed \"%s\": not owned by UID %u",
-                path, (unsigned) geteuid ());
+        log_msg (LOG_WARNING, "Ignoring PRNG seed \"%s\": must be owned by "
+                "UID %u instead of UID %u", path, (unsigned) geteuid (),
+                (unsigned) st.st_uid);
     }
-    else if (st.st_mode & (S_IRGRP | S_IROTH)) {
+    else if (st.st_mode & (S_IRGRP | S_IWGRP)) {
         log_msg (LOG_WARNING,
-                "Ignoring PRNG seed \"%s\": readable by group or other", path);
+                "Ignoring PRNG seed \"%s\": must not be readable or writable "
+                "by group (perms=%04o)", path, (st.st_mode & ~S_IFMT));
     }
-    else if (st.st_mode & (S_IWGRP | S_IWOTH)) {
+    else if (st.st_mode & (S_IROTH | S_IWOTH)) {
         log_msg (LOG_WARNING,
-                "Ignoring PRNG seed \"%s\": writable by group or other", path);
+                "Ignoring PRNG seed \"%s\": must not be readable or writable "
+                "by other (perms=%04o)", path, (st.st_mode & ~S_IFMT));
     }
     else {
         is_valid = 1;
@@ -509,7 +416,7 @@ retry_open:
     }
     n = num_bytes - num_left;
     if (n > 0) {
-        log_msg (LOG_INFO, "PRNG seeded with %d byte%s from \"%s\"",
+        log_msg (LOG_INFO, "Seeded PRNG with %d byte%s from \"%s\"",
                 n, (n == 1 ? "" : "s"), path);
     }
     return (!is_valid ? -1 : n);
@@ -607,42 +514,31 @@ _random_stir_entropy (void *_arg_not_used_)
 {
 /*  Periodically stirs the entropy pool by mixing in new entropy.
  */
-    static uint32_t cnt;
-    clock_t         cpu;
-    struct timeval  tv;
-    int             msecs;
+    unsigned buf;
+    int      msecs;
 
     assert (RANDOM_STIR_MAX_SECS > 0);
 
     if (_random_stir_secs <= 0) {
         return;
     }
-    if (cnt == 0) {
-        _random_bytes (&cnt, sizeof (cnt));
-    }
     _random_timer_id = 0;
 
     log_msg (LOG_DEBUG, "Stirring PRNG entropy pool");
 
-    cpu = clock ();
-    if (cpu != (clock_t) -1) {
-        cnt += (uint32_t) cpu;
+    if (entropy_read_uint (&buf) != -1) {
+        _random_add (&buf, sizeof (buf));
     }
-    if (gettimeofday (&tv, NULL) == 0) {
-        cnt += (uint32_t) (tv.tv_sec + tv.tv_usec);
-    }
-    _random_add (&cnt, sizeof (cnt));
-
     /*  Perform an exponential backoff up to the maximum timeout.  This allows
      *    for vigorous stirring of the entropy pool when the daemon is started.
      */
     if (_random_stir_secs < RANDOM_STIR_MAX_SECS) {
         _random_stir_secs = MIN(_random_stir_secs * 2, RANDOM_STIR_MAX_SECS);
     }
-    /*  The 10 low-order bits of the current time are used to mix things up and
-     *    stagger subsequent timer callbacks by up to 1023ms.
+    /*  The 10 low-order bits are used to stagger subsequent timer callbacks
+     *    by up to 1023ms.
      */
-    msecs = (_random_stir_secs * 1000) + (tv.tv_usec & 0x3FF);
+    msecs = (_random_stir_secs * 1000) + (buf & 0x3FF);
 
     _random_timer_id = timer_set_relative (
             (callback_f) _random_stir_entropy, NULL, msecs);
